from sapextractor.algo.p2p.tab_processing import eban_processing, ekko_processing, ekpo_processing, rbkp_processing, rseg_processing, ekbe_processing, bkpf_processing
import networkx as nx
import pandas as pd
from dateutil import parser


def add_edges_to_graph(edges, nodes_connections, G):
    for edge in edges:
        if edge[0] in G.nodes and edge[1] in G.nodes:
            G.add_edge(edge[0], edge[1])
        if not edge[0] in nodes_connections:
            nodes_connections[edge[0]] = {edge[0]}
        if not edge[1] in nodes_connections:
            nodes_connections[edge[1]] = {edge[1]}
        nodes_connections[edge[0]].add(edge[1])
    return nodes_connections, G


def extract_tables_and_graph(con, gjahr=None, min_extr_date=None):
    G = nx.DiGraph()
    nodes_types = {}
    nodes_connections = {}
    eban, eban_nodes_types = eban_processing.apply(con)
    for n in eban_nodes_types:
        G.add_node(n)
    nodes_types.update(eban_nodes_types)
    ekko, ekko_nodes_types = ekko_processing.apply(con)
    for n in ekko_nodes_types:
        G.add_node(n)
    nodes_types.update(ekko_nodes_types)
    eban_ekko_connection = ekpo_processing.eban_ekko_connection(con)
    nodes_connections, G = add_edges_to_graph(eban_ekko_connection, nodes_connections, G)

    rbkp, rbkp_nodes_types = rbkp_processing.apply(con, gjahr=gjahr)
    for n in rbkp_nodes_types:
        G.add_node(n)
    nodes_types.update(rbkp_nodes_types)
    ekko_rbkp_connections = rseg_processing.apply(con, gjahr=gjahr)
    nodes_connections, G = add_edges_to_graph(ekko_rbkp_connections, nodes_connections, G)

    gr, gr_nodes_types = ekbe_processing.goods_receipt(con, gjahr=gjahr)
    for n in gr_nodes_types:
        G.add_node(n)
    nodes_types.update(gr_nodes_types)
    gr_ekko_connections = ekbe_processing.goods_receipt_ekko_connection(gr)
    nodes_connections, G = add_edges_to_graph(gr_ekko_connections, nodes_connections, G)

    ir, ir_nodes_types = ekbe_processing.invoice_receipt(con, gjahr=gjahr)
    for n in ir_nodes_types:
        G.add_node(n)
    nodes_types.update(ir_nodes_types)
    ir_ekko_connections = ekbe_processing.invoice_receipt_ekko_connection(ir)
    nodes_connections, G = add_edges_to_graph(ir_ekko_connections, nodes_connections, G)

    bkpf_events, bkpf_doc_types, bkpf_connections = bkpf_processing.apply(con, gjahr=gjahr)
    for n in bkpf_doc_types:
        G.add_node(n)
    nodes_types.update(bkpf_doc_types)
    nodes_connections, G = add_edges_to_graph(bkpf_connections, nodes_connections, G)

    nodes_connections = pd.DataFrame([{"node": x, "RELATED_DOCUMENTS": list(y)} for x, y in nodes_connections.items()])
    dataframe = pd.concat([eban, ekko, rbkp, gr, ir, bkpf_events])
    dataframe = dataframe.sort_values("event_timestamp")
    dataframe = dataframe.merge(nodes_connections, left_on="event_node", right_on="node", suffixes=('', '_r'), how="left")

    if min_extr_date is not None:
        min_extr_date = parser.parse(min_extr_date)
        dataframe = dataframe[dataframe["event_timestamp"] >= min_extr_date]

    return dataframe, G, nodes_types



from sapextractor.algo.p2p import p2p_1d_dataframe, p2p_1d_log
from sapextractor.algo.p2p import obj_centr_log


def cli(con):
    print("\n\nP2P extraction\n")
    print("available extraction types:")
    print("1) P2P classic log (XES log)")
    print("2) P2P classic log (dataframe)")
    print("3) P2P object-centric log")
    print()
    ext_type = input("insert your choice (default: 1): ")
    if not ext_type:
        ext_type = "1"
    if ext_type == "1":
        return p2p_1d_log.cli(con)
    elif ext_type == "2":
        return p2p_1d_dataframe.cli(con)
    elif ext_type == "3":
        return obj_centr_log.cli(con)




from sapextractor.algo.ap_ar import single_doc_transactions_dataframe
from pm4py.objects.conversion.log import converter as log_converter
from pm4py.objects.log.util import sorting
from pm4py.objects.log.exporter.xes import exporter as xes_exporter


def apply(con, gjahr="1997", bukrs=None, **ext_arg):
    dataframe = single_doc_transactions_dataframe.apply(con, gjahr=gjahr, bukrs=bukrs)
    log = log_converter.apply(dataframe, parameters={"stream_postprocessing": True})
    log = sorting.sort_timestamp(log, "time:timestamp")
    return log


def cli(con):
    print("\n\nAccounting - Transactions for the single document (XES log)\n")
    log = apply(con)
    path = input("Insert the path where the log should be saved (default: bkpf.xes):")
    if not path:
        path = "bkpf.xes"
    xes_exporter.apply(log, path)



from sapextractor.algo.o2c import o2c_1d_log_extractor, o2c_1d_dataframe_extractor
from sapextractor.algo.o2c import obj_centr_log


def apply(con, ext_type, ext_arg):
    if ext_type == "o2c_1d_log_extractor":
        return o2c_1d_log_extractor.apply(con, **ext_arg)
    elif ext_type == "o2c_1d_dataframe_extractor":
        return o2c_1d_dataframe_extractor.apply(con, **ext_arg)
    elif ext_type == "o2c_obj_centric_log":
        return obj_centr_log.apply(con, **ext_arg)



from sapextractor.database_connection.interface import DatabaseConnection
from sapextractor.utils.string_matching import find_corr
import sqlite3
import pandas as pd
from sapextractor.utils import constants


class SqliteConnection(DatabaseConnection):
    def __init__(self, path):
        self.path = path
        self.con = sqlite3.connect(self.path)
        constants.TIMESTAMP_FORMAT = "%d.%m.%Y %H:%M:%S"
        self.table_prefix = ""
        DatabaseConnection.__init__(self)

    def execute_read_sql(self, sql, columns):
        cursor = self.con.cursor()
        cursor.execute(sql)
        stream = []
        df = []
        while True:
            res = cursor.fetchmany(10000)
            if len(res) == 0:
                break
            for row in res:
                el = {}
                for idx, col in enumerate(columns):
                    el[col] = row[idx]
                stream.append(el)
            this_dataframe = pd.DataFrame(stream)
            df.append(this_dataframe)
            stream = None
            stream = []
        if df:
            df = pd.concat(df)
        else:
            df = pd.DataFrame({x: [] for x in columns})
        df.columns = [x.upper() for x in df.columns]
        return df

    def get_list_tables(self):
        cursor = self.con.cursor()
        cursor.execute('SELECT name from sqlite_master where type= "table"')
        tables = cursor.fetchall()
        tables = [x[0] for x in tables]
        return tables

    def write_dataframe(self, dataframe, table_name):
        dataframe.to_sql(table_name, con=self.con)

    def get_columns(self, table_name):
        cursor = self.con.cursor()
        cursor.execute("SELECT name FROM PRAGMA_TABLE_INFO('%s')" % table_name)
        columns = cursor.fetchall()
        columns = [x[0] for x in columns]
        return columns

    def format_table_name(self, table_name):
        return table_name

    def prepare_query(self, table_name, columns):
        table_name = self.format_table_name(table_name)
        table_columns = self.get_columns(table_name)
        columns = find_corr.apply(columns, table_columns)
        return "SELECT "+",".join(columns)+" FROM "+table_name

    def prepare_and_execute_query(self, table_name, columns, additional_query_part=""):
        query = self.prepare_query(table_name, columns) + additional_query_part
        dataframe = self.execute_read_sql(query, columns)
        dataframe.columns = columns
        return dataframe


def apply(path='./sap.sqlite'):
    return SqliteConnection(path)


def cli():
    print("\n\n")
    print("== Connection to a SQLite database == \n\n")
    path = input("Insert the path to the SQLite database (default: ./sap.sqlite):")
    if not path:
        path = "./sap.sqlite"
    con = apply(path)
    return con



from sapextractor.algo.ap_ar import cli, ap_ar_common, single_doc_transactions_dataframe, single_doc_transactions_log
from sapextractor.algo.ap_ar import document_flow_dataframe, document_flow_log, factory
from sapextractor.algo.ap_ar import doc_flow_transactions_dataframe, doc_flow_transactions_log
from sapextractor.algo.ap_ar import obj_centr_log



import pandas as pd


def apply(con):
    eban = con.prepare_and_execute_query("EBAN", ["BANFN", "ERNAM", "ERDAT"])
    eban.columns = ["event_"+x for x in eban.columns]
    eban = eban.rename(columns={"event_ERNAM": "event_USERNAME", "event_ERDAT": "event_timestamp"})
    eban = eban.sort_values("event_timestamp")
    eban = eban.groupby("event_BANFN").first().reset_index()
    eban["event_timestamp"] = pd.to_datetime(eban["event_timestamp"])
    eban["event_activity"] = "Create Purchase Requisition"
    eban["event_FROMTABLE"] = "EBAN"
    eban["event_node"] = "EBAN_"+eban["event_BANFN"]
    eban = eban.dropna(subset=["event_node"])
    eban_nodes_types = {x: "EBAN" for x in eban["event_node"].unique()}
    return eban, eban_nodes_types



from sapextractor.database_connection import factory as conn_factory
from sapextractor.algo import factory as algo_factory


def apply(db_type, db_con_arg, process, ext_type, ext_arg):
    con = conn_factory.apply(db_type, db_con_arg)
    return algo_factory.apply(con, process, ext_type, ext_arg)



def eban_ekko_connection(con):
    ekpo = con.prepare_and_execute_query("EKPO", ["EBELN", "BANFN"])
    ekpo = ekpo.dropna(subset=["EBELN", "BANFN"], how="any")
    ekpo = ekpo.to_dict('records')
    ekpo = [("EBAN_"+x["BANFN"], "EKKO_"+x["EBELN"]) for x in ekpo]
    return ekpo



from sapextractor.algo.ap_ar import factory as ap_ar_factory
from sapextractor.algo.o2c import factory as o2c_factory
from sapextractor.algo.p2p import factory as p2p_factory


def apply(con, process, ext_type, ext_arg):
    if process == "ap_ar":
        return ap_ar_factory.apply(con, ext_type, ext_arg)
    elif process == "o2c":
        return o2c_factory.apply(con, ext_type, ext_arg)
    elif process == "p2p":
        return p2p_factory.apply(con, ext_type, ext_arg)



import pandas as pd


def apply(con, gjahr=None):
    additional_query_part = ""
    if gjahr is not None:
        additional_query_part = " WHERE GJAHR = '"+gjahr+"' AND AUGGJ = '"+gjahr+"'"
    bsak = con.prepare_and_execute_query("BSAK", ["BELNR", "BUZEI", "GJAHR", "AUGBL", "AUGGJ"], additional_query_part=additional_query_part)
    bsak = bsak.to_dict("r")
    clearances = {}
    for x in bsak:
        key = x["BELNR"]+x["GJAHR"]
        value = x["AUGBL"]+x["AUGGJ"]
        if not key in clearances:
            clearances[key] = list()
        clearances[key].append(value)
    return clearances



import pandas as pd


def apply(con, gjahr=None):
    additional_query_part = ""
    if gjahr is not None:
        additional_query_part = " WHERE GJAHR = '"+gjahr+"'"
    rseg = con.prepare_and_execute_query("RSEG", ["BELNR", "GJAHR", "EBELN"], additional_query_part=additional_query_part)
    rseg = rseg.dropna(subset=["EBELN"])
    rseg = rseg.to_dict("records")
    rseg = list(set(("EKKO_"+x["EBELN"], "RBKP_"+x["BELNR"]+x["GJAHR"]) for x in rseg))
    return rseg



DOC_SEP = "@#@#@#"
MAX_CASE_SIZE = 130

ORACLE_OWNER = "SAPSR3"
ORACLE_ARRAYSIZE = 10000

TIMESTAMP_FORMAT = "%d.%m.%Y %H:%M:%S"


def set_documents(content):
    ret = []
    content = content.split(DOC_SEP)
    for x in content:
        if str(x).lower() != "nan":
            ret.append(x)
    return ret


INVOLVED_TABLES = [
    "T003T",
    "CDHDR",
    "CDPOS",
    "TSTCT",
    "DD07T",
    "BKPF",
    "BSEG",
    "VBFA",
    "EBAN",
    "EKKO",
    "EKPO",
    "MKPF",
    "MSEG"
]



import pandas as pd

from sapextractor.utils.dates import timestamp_column_from_dt_tm
from sapextractor.utils.tstct import extract_tstct
from sapextractor.utils.fields_corresp import extract_dd03t
from sapextractor.utils.change_tables import mapping


def read_cdhdr(con, objectclas=None):
    additional_query_part = " WHERE OBJECTCLAS = '" + objectclas + "'" if objectclas is not None else ""
    df = con.prepare_and_execute_query("CDHDR", ["CHANGENR", "USERNAME", "UDATE", "UTIME", "TCODE"],
                                       additional_query_part=additional_query_part)
    df.columns = ["event_" + x for x in df.columns]
    if len(df) > 0:
        df = timestamp_column_from_dt_tm.apply(df, "event_UDATE", "event_UTIME", "event_timestamp")
        df = df.sort_values("event_timestamp")
    transactions = set(df["event_TCODE"].unique())
    tstct = extract_tstct.apply_static(con, transactions=transactions)
    df["event_ONLYACT"] = df["event_TCODE"].map(tstct)
    return df


def read_cdpos(con, objectclas=None, tabname=None):
    additional_query_part = " WHERE OBJECTCLAS = '" + objectclas + "'" if objectclas is not None else ""
    if tabname is not None and additional_query_part:
        additional_query_part += " AND TABNAME = '" + tabname + "'"
    df = con.prepare_and_execute_query("CDPOS", ["CHANGENR", "OBJECTID", "TABNAME", "FNAME", "VALUE_NEW", "VALUE_OLD", "CHNGIND"],
                                       additional_query_part=additional_query_part)
    return df


def give_field_desc(con, cdpos_dict):
    #fnames = extract_dd03t.apply(con)
    fnames = {}
    for c in cdpos_dict:
        fname = c["FNAME"]
        tabname = c["TABNAME"]
        value_new = c["VALUE_NEW"]
        value_old = c["VALUE_OLD"]
        chngind = c["CHNGIND"]

        change, typ = mapping.perform_mapping(tabname, fname, value_old, value_new, chngind, fnames)
        c["CHANGEDESC"] = change


    return cdpos_dict


def apply(con, objectclas=None, tabname=None):
    cdhdr = read_cdhdr(con, objectclas=objectclas)
    cdpos = read_cdpos(con, objectclas=objectclas, tabname=tabname)
    grouped_cdhdr = cdhdr.groupby("event_CHANGENR")
    change_dictio = {}
    for name, group in grouped_cdhdr:
        change_dictio[name] = group
    del grouped_cdhdr
    cdpos = cdpos.to_dict('records')
    cdpos = give_field_desc(con, cdpos)
    ret = {}
    for el in cdpos:
        changenr = el["CHANGENR"]
        objectid = el["OBJECTID"]
        tabname = el["TABNAME"]
        fname = el["FNAME"]
        value_new = el["VALUE_NEW"]
        changedesc = el["CHANGEDESC"]
        chngind = el["CHNGIND"]
        if changenr in change_dictio:
            if objectid not in ret:
                ret[objectid] = []
            df = change_dictio[changenr].copy()
            df["event_AWKEY"] = objectid
            df["event_TABNAME"] = tabname
            df["event_FNAME"] = fname
            df["event_VALUE_NEW"] = value_new
            df["event_CHANGEDESC"] = changedesc
            df["event_CHNGIND"] = chngind
            ret[objectid].append(df)
    for objectid in ret:
        ret[objectid] = pd.concat(ret[objectid])
    return ret


def get_changes_dataframe_two_mapping(con, orig_df, change_key, target_key, resource_column=None, objectclas=None,
                                      tabname=None):
    changes = apply(con, objectclas=objectclas, tabname=tabname)
    mapping = {x[change_key]: x[target_key] for x in
               orig_df[[change_key, target_key]].dropna(subset=[change_key, target_key], how="any").to_dict("records")}
    changes = {x: y for x, y in changes.items() if x in mapping}
    changes = [y for y in changes.values()]
    if changes:
        changes = pd.concat(changes)
        changes[target_key] = changes[change_key].map(mapping)
        changes = changes.dropna(subset=[change_key, target_key, "event_ONLYACT"], how="any")
        if resource_column is not None:
            changes = changes.rename(columns={"event_USERNAME": resource_column})
    else:
        changes = pd.DataFrame()
    return changes



import pandas as pd


def goods_receipt(con, gjahr=None):
    additional_query_part = " WHERE VGABE = '1'"
    if gjahr is not None:
        additional_query_part += " AND GJAHR = '"+gjahr+"'"
    ekbe = con.prepare_and_execute_query("EKBE", ["EBELN", "EBELP", "BELNR", "BUZEI", "BUDAT", "GJAHR"], additional_query_part=additional_query_part)
    ekbe.columns = ["event_"+x for x in ekbe.columns]
    ekbe["event_timestamp"] = pd.to_datetime(ekbe["event_BUDAT"], errors="coerce")
    ekbe = ekbe.dropna(subset=["event_timestamp"])
    ekbe["event_FROMTABLE"] = "EKBE"
    ekbe["event_node"] = "EKBEGR_"+ekbe["event_BELNR"]+ekbe["event_GJAHR"]
    ekbe["event_activity"] = "Goods Receipt"
    ekbe["event_timestamp"] = ekbe["event_timestamp"] + pd.Timedelta("1 second")
    ekbe_nodes_types = {x: "EKBEGR" for x in ekbe["event_node"].unique()}
    return ekbe, ekbe_nodes_types


def invoice_receipt(con, gjahr=None):
    additional_query_part = " WHERE VGABE = '2'"
    if gjahr is not None:
        additional_query_part += " AND GJAHR = '"+gjahr+"'"
    ekbe = con.prepare_and_execute_query("EKBE", ["EBELN", "EBELP", "BELNR", "BUZEI", "BUDAT", "GJAHR"], additional_query_part=additional_query_part)
    ekbe.columns = ["event_"+x for x in ekbe.columns]
    ekbe["event_timestamp"] = pd.to_datetime(ekbe["event_BUDAT"], errors="coerce")
    ekbe = ekbe.dropna(subset=["event_timestamp"])
    ekbe["event_FROMTABLE"] = "EKBE"
    ekbe["event_node"] = "EKBEIR_"+ekbe["event_BELNR"]+ekbe["event_GJAHR"]
    ekbe["event_activity"] = "Invoice Receipt"
    ekbe["event_timestamp"] = ekbe["event_timestamp"] + pd.Timedelta("2 seconds")
    ekbe_nodes_types = {x: "EKBEIR" for x in ekbe["event_node"].unique()}
    return ekbe, ekbe_nodes_types


def goods_receipt_ekko_connection(gr):
    gr = gr[["event_EBELN", "event_BELNR", "event_GJAHR"]].to_dict("r")
    cons = list(set(("EKKO_"+x["event_EBELN"], "EKBEGR_"+x["event_BELNR"]+x["event_GJAHR"]) for x in gr))
    return cons


def invoice_receipt_ekko_connection(ir):
    ir = ir[["event_EBELN", "event_BELNR", "event_GJAHR"]].to_dict("r")
    cons = list(set(("EKKO_"+x["event_EBELN"], "EKBEIR_"+x["event_BELNR"]+x["event_GJAHR"]) for x in ir))
    return cons



import pandas as pd


def apply(con, gjahr=None):
    additional_query_part = ""
    if gjahr is not None:
        additional_query_part = " WHERE GJAHR = '"+gjahr+"'"
    rbkp = con.prepare_and_execute_query("RBKP", ["BELNR", "GJAHR", "BLDAT", "BUDAT", "USNAME", "TCODE", "LIFNR"], additional_query_part=additional_query_part)
    rbkp.columns = ["event_"+x for x in rbkp.columns]
    rbkp["event_FROMTABLE"] = "RBKP"
    rbkp["event_node"] = "RBKP_"+rbkp["event_BELNR"]+rbkp["event_GJAHR"]
    rbkp1 = rbkp.copy()
    rbkp1["event_timestamp"] = pd.to_datetime(rbkp1["event_BLDAT"], errors="coerce")
    rbkp1["event_activity"] = "Invoice Emitted"
    rbkp1 = rbkp1.dropna(subset=["event_timestamp"])
    rbkp2 = rbkp.copy()
    rbkp2["event_timestamp"] = pd.to_datetime(rbkp1["event_BUDAT"], errors="coerce")
    rbkp2["event_activity"] = "Invoice Posted"
    rbkp2 = rbkp2.dropna(subset=["event_timestamp"])
    rbkp2["event_timestamp"] = rbkp2["event_timestamp"] + pd.Timedelta("3 seconds")
    rbkp = pd.concat([rbkp1, rbkp2])
    rbkp_nodes_types = {x: "RBKP" for x in rbkp["event_node"].unique()}
    return rbkp, rbkp_nodes_types



import pandas as pd
from frozendict import frozendict
from sapextractor.algo.p2p.tab_processing import bsak_processing


def apply(con, gjahr=None):
    awkey_docs, doc_dates, doc_types = extract_docs_from_bkpf(con, gjahr=gjahr)
    clearances = bsak_processing.apply(con, gjahr=gjahr)
    events, doc_types, connections = mix_bkpf_bseg(awkey_docs, doc_dates, doc_types, clearances)
    return events, doc_types, connections


def extract_docs_from_bkpf(con, gjahr=None):
    additional_query_part = ""
    if gjahr is not None:
        additional_query_part = " WHERE GJAHR = '"+gjahr+"'"
    bkpf = con.prepare_and_execute_query("BKPF", ["BELNR", "GJAHR", "BLART", "BLDAT", "AWKEY"], additional_query_part=additional_query_part)
    bkpf["BLDAT"] = pd.to_datetime(bkpf["BLDAT"], errors="coerce") + pd.Timedelta("6 seconds")
    bkpf = bkpf.dropna(subset=["BLDAT"])
    bkpf = bkpf.to_dict("r")
    awkey_docs = {}
    doc_dates = {}
    doc_types = {}
    for el in bkpf:
        doc_dates[el["BELNR"] + el["GJAHR"]] = el["BLDAT"]
        doc_types[el["BELNR"] + el["GJAHR"]] = el["BLART"]
        try:
            year = el["AWKEY"][-4:]
            docnum = el["AWKEY"][:10]
            key = docnum+year
            if key not in awkey_docs:
                awkey_docs[key] = list()
            awkey_docs[key].append(el["BELNR"]+el["GJAHR"])
        except:
            pass
    return awkey_docs, doc_dates, doc_types


def mix_bkpf_bseg(awkey_docs, doc_dates, doc_types, clearances):
    events = set()
    nodes_types = {}
    connections = set()

    for inv in awkey_docs:
        for docitem in awkey_docs[inv]:
            if docitem in clearances:
                for clear in clearances[docitem]:
                    if clear in doc_types and clear in doc_dates:
                        doc_type = doc_types[clear]
                        doc_date = doc_dates[clear]
                        events.add(frozendict({"event_node": "BKPF_"+clear, "event_timestamp": doc_date, "event_activity": "Clearance ("+doc_type+")", "event_FROMTABLE": "BKPF"}))
                        nodes_types["BKPF_"+clear] = "BKPF"
                        connections.add(("RBKP_"+inv, "BKPF_"+clear))
    events = list(events)
    for i in range(len(events)):
        events[i] = dict(events[i])
    events = pd.DataFrame(events)

    return events, nodes_types, connections



from sapextractor.algo.p2p import p2p_common
from pm4pymdl.objects.ocel.exporter import exporter as jmd_exporter
from pm4pymdl.objects.mdl.exporter import exporter as mdl_exporter


def apply(con, gjahr="2014", min_extr_date="2014-01-01 00:00:00"):
    dataframe, G, nodes_types = p2p_common.extract_tables_and_graph(con, gjahr=gjahr, min_extr_date=min_extr_date)
    dataframe["event_id"] = dataframe.index.astype(str)
    dataframe = dataframe.sort_values("event_timestamp")
    dataframe.type = "succint"
    return dataframe


def cli(con):
    print("\n\nP2P - Object-Centric Log\n")
    dataframe = apply(con)
    path = input("Insert the path where the log should be saved (default: p2p.mdl): ")
    if not path:
        path = "p2p.xmlocel"
    if path.endswith("mdl"):
        mdl_exporter.apply(dataframe, path)
    elif path.endswith("jsonocel") or path.endswith("xmlocel"):
        jmd_exporter.apply(dataframe, path)



from sapextractor.database_connection import oracle, sqlite


def apply(db_type, db_con_args):
    if db_type == "sqlite":
        return sqlite.apply(**db_con_args)
    elif db_type == "oracle":
        return oracle.apply(**db_con_args)



from sapextractor.utils.vbtyp import extract_vbtyp


def extract_dfg(con):
    vbfa = con.execute_read_sql("SELECT VBTYP_V, VBTYP_N, Count(*) FROM "+con.table_prefix+"VBFA GROUP BY VBTYP_V, VBTYP_N", ["VBTYP_V", "VBTYP_N", "COUNT"])
    doc_types = set(vbfa["VBTYP_N"].unique()).union(set(vbfa["VBTYP_V"].unique()))
    vbtyp = extract_vbtyp.apply_static(con, doc_types=doc_types)
    vbfa["VBTYP_V"] = vbfa["VBTYP_V"].map(vbtyp)
    vbfa["VBTYP_N"] = vbfa["VBTYP_N"].map(vbtyp)
    vbfa = vbfa.dropna(subset=["VBTYP_V", "VBTYP_N"], how="any")
    vbfa = vbfa.to_dict("r")
    dfg = {(x["VBTYP_V"], x["VBTYP_N"]): int(x["COUNT"]) for x in vbfa}
    start_activities = con.execute_read_sql("SELECT VBTYP_V, Count(*) FROM (SELECT VBELV, VBTYP_V FROM "+con.table_prefix+"VBFA WHERE VBELV NOT IN (SELECT VBELN FROM "+con.table_prefix+"VBFA)) GROUP BY VBTYP_V", ["ACT", "COUNT"])
    end_activities = con.execute_read_sql("SELECT VBTYP_N, Count(*) FROM (SELECT VBELN, VBTYP_N FROM "+con.table_prefix+"VBFA WHERE VBELN NOT IN (SELECT VBELV FROM "+con.table_prefix+"VBFA)) GROUP BY VBTYP_N", ["ACT", "COUNT"])
    start_activities["ACT"] = start_activities["ACT"].map(vbtyp)
    end_activities["ACT"] = end_activities["ACT"].map(vbtyp)
    start_activities = start_activities.to_dict("r")
    end_activities = end_activities.to_dict("r")
    start_activities = {x["ACT"]: x["COUNT"] for x in start_activities}
    end_activities = {x["ACT"]: x["COUNT"] for x in end_activities}
    act_count_exit = {}
    for el in dfg:
        if not el[0] in act_count_exit:
            act_count_exit[el[0]] = 0
        act_count_exit[el[0]] += int(dfg[el])
    act_count_entry = {}
    for el in dfg:
        if not el[1] in act_count_entry:
            act_count_entry[el[1]] = 0
        act_count_entry[el[1]] += int(dfg[el])
    act_count = {}
    activities = set(act_count_entry).union(set(act_count_exit))
    for act in activities:
        if act in act_count_entry and act in act_count_exit:
            act_count[act] = max(act_count_entry[act], act_count_exit[act])
        elif act in act_count_entry:
            act_count[act] = act_count_entry[act]
        elif act in act_count_exit:
            act_count[act] = act_count_exit[act]
    return dfg, act_count, start_activities, end_activities